@@ -1,4 +1,4 @@
-/* 
+/*
  * Copyright (C) 2006, 2007 OpenedHand Ltd.
  *
  * Author: Jorn Baayen <jorn@openedhand.com>
@@ -42,6 +42,7 @@
 #include "xml-util.h"
 #include "gena-protocol.h"
 #include "accept-language.h"
+#include "gvalue-util.h"
 
 G_DEFINE_TYPE (GUPnPServiceProxy,
                gupnp_service_proxy,
@@ -60,6 +61,8 @@ struct _GUPnPServiceProxyPrivate {
         int seq; /* Event sequence number */
 
         GHashTable *notify_hash;
+
+        GList *pending_messages; /* Pending SoupMessages from this proxy */
 };
 
 enum {
@@ -103,7 +106,7 @@ subscribe_got_response (SoupMessage       *msg,
 static void
 subscribe (GUPnPServiceProxy *proxy);
 static void
-unsubscribe (GUPnPServiceProxy *proxy, gboolean sync);
+unsubscribe (GUPnPServiceProxy *proxy);
 
 static void
 gupnp_service_proxy_action_free (GUPnPServiceProxyAction *action)
@@ -192,18 +195,47 @@ gupnp_service_proxy_dispose (GObject *object)
 {
         GUPnPServiceProxy *proxy;
         GObjectClass *object_class;
+        GUPnPContext *context;
+        SoupSession *session;
 
         proxy = GUPNP_SERVICE_PROXY (object);
 
-        /* Cancel any pending actions */
+        /* Unsubscribe */
+        if (proxy->priv->subscribed) {
+                unsubscribe (proxy);
+
+                proxy->priv->subscribed = FALSE;
+        }
+
+        /* Cancel pending actions */
         while (proxy->priv->pending_actions) {
                 GUPnPServiceProxyAction *action;
 
                 action = proxy->priv->pending_actions->data;
-                
+
                 gupnp_service_proxy_cancel_action (proxy, action);
         }
 
+        /* Cancel pending messages */
+        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (proxy));
+        if (context)
+                session = _gupnp_context_get_session (context);
+        else
+                session = NULL; /* Not the first time dispose is called. */
+
+        while (proxy->priv->pending_messages) {
+                SoupMessage *msg;
+
+                msg = proxy->priv->pending_messages->data;
+
+                soup_message_set_status (msg, SOUP_STATUS_CANCELLED);
+                soup_session_cancel_message (session, msg);
+
+                proxy->priv->pending_messages =
+                        g_list_delete_link (proxy->priv->pending_messages,
+                                            proxy->priv->pending_messages);
+        }
+
         /* Call super */
         object_class = G_OBJECT_CLASS (gupnp_service_proxy_parent_class);
         object_class->dispose (object);
@@ -257,7 +289,7 @@ gupnp_service_proxy_class_init (GUPnPServiceProxyClass *klass)
                                        G_PARAM_STATIC_NAME |
                                        G_PARAM_STATIC_NICK |
                                        G_PARAM_STATIC_BLURB));
-                                       
+
         /**
          * GUPnPServiceProxy::subscription-lost
          * @proxy: The #GUPnPServiceProxy that received the signal
@@ -347,7 +379,7 @@ gupnp_service_proxy_send_action_valist (GUPnPServiceProxy *proxy,
         GMainContext *main_context;
         GMainLoop *main_loop;
         GUPnPServiceProxyAction *handle;
-        
+
         g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), FALSE);
         g_return_val_if_fail (action, FALSE);
 
@@ -382,6 +414,66 @@ gupnp_service_proxy_send_action_valist (GUPnPServiceProxy *proxy,
         return TRUE;
 }
 
+/**
+ * gupnp_service_proxy_send_action_hash
+ * @proxy: A #GUPnPServiceProxy
+ * @action: An action
+ * @error: The location where to store any error, or NULL
+ * @in_hash: A #GHashTable of in parameter name and #GValue pairs
+ * @out_hash: A #GHashTable of out parameter name and initialized
+ * #GValue pairs
+ *
+ * See gupnp_service_proxy_send_action(); this version takes a pair of
+ * #GHashTable<!-- -->s for runtime determined parameter lists.
+ *
+ * Return value: TRUE if sending the action was succesful.
+ **/
+gboolean
+gupnp_service_proxy_send_action_hash (GUPnPServiceProxy *proxy,
+                                      const char        *action,
+                                      GError           **error,
+                                      GHashTable        *in_hash,
+                                      GHashTable        *out_hash)
+{
+        GUPnPContext *context;
+        GMainContext *main_context;
+        GMainLoop *main_loop;
+        GUPnPServiceProxyAction *handle;
+
+        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), FALSE);
+        g_return_val_if_fail (action, FALSE);
+
+        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (proxy));
+        main_context = gssdp_client_get_main_context (GSSDP_CLIENT (context));
+        main_loop = g_main_loop_new (main_context, TRUE);
+
+        handle = gupnp_service_proxy_begin_action_hash (proxy,
+                                                        action,
+                                                        stop_main_loop,
+                                                        main_loop,
+                                                        error,
+                                                        in_hash);
+        if (!handle) {
+                g_main_loop_unref (main_loop);
+
+                return FALSE;
+        }
+
+        /* Loop till we get a reply (or time out) */
+        if (g_main_loop_is_running (main_loop))
+                g_main_loop_run (main_loop);
+
+        g_main_loop_unref (main_loop);
+
+        if (!gupnp_service_proxy_end_action_hash (proxy,
+                                                  handle,
+                                                  error,
+                                                  out_hash))
+                return FALSE;
+
+        return TRUE;
+}
+
 /**
  * gupnp_service_proxy_begin_action
  * @proxy: A #GUPnPServiceProxy
@@ -424,82 +516,15 @@ gupnp_service_proxy_begin_action (GUPnPServiceProxy              *proxy,
         return ret;
 }
 
-/* Received response to action message */
-static void
-action_got_response (SoupMessage             *msg,
-                     GUPnPServiceProxyAction *action)
+/* Begins a basic action message */
+static SoupSoapMessage *
+begin_action_msg (GUPnPServiceProxy *proxy,
+                  const char        *action,
+                  GError           **error)
 {
-        if (msg->status == SOUP_STATUS_METHOD_NOT_ALLOWED) {
-                /* Not allowed .. */
-                const char *full_action;
-                GUPnPContext *context;
-                SoupSession *session;
-
-                /* Retry with M-POST */
-                msg->method = "M-POST";
-
-                soup_message_add_header
-                        (msg->request_headers,
-                         "Man",
-                         "\"http://schemas.xmlsoap.org/soap/envelope/\"; ns=s");
-
-                /* Rename "SOAPAction" to "s-SOAPAction" */
-                full_action = soup_message_get_header (msg->request_headers,
-                                                       "SOAPAction");
-                soup_message_add_header (msg->request_headers,
-                                         "s-SOAPAction",
-                                         full_action);
-                soup_message_remove_header (msg->request_headers,
-                                            "SOAPAction");
-
-                /* And re-queue */
-                context = gupnp_service_info_get_context
-                                (GUPNP_SERVICE_INFO (action->proxy));
-                session = _gupnp_context_get_session (context);
-
-                soup_session_requeue_message (session, msg);
-        } else {
-                /* Success: Call callback */
-                action->callback (action->proxy, action, action->user_data);
-        }
-}
-
-/**
- * gupnp_service_proxy_begin_action_valist
- * @proxy: A #GUPnPServiceProxy
- * @action: An action
- * @callback: The callback to call when sending the action has succeeded
- * or failed
- * @user_data: User data for @callback
- * @error: The location where to store any error, or NULL
- * @var_args: A va_list of tuples of in parameter name, in paramater type, and 
- * in parameter value
- *
- * See gupnp_service_proxy_begin_action(); this version takes a va_list for
- * use by language bindings.
- *
- * Return value: A #GUPnPServiceProxyAction handle, or NULL on error. This will
- * be freed automatically on @callback calling.
- **/
-GUPnPServiceProxyAction *
-gupnp_service_proxy_begin_action_valist
-                                   (GUPnPServiceProxy              *proxy,
-                                    const char                     *action,
-                                    GUPnPServiceProxyActionCallback callback,
-                                    gpointer                        user_data,
-                                    GError                        **error,
-                                    va_list                         var_args)
-{
-        char *control_url, *full_action, *lang;
-        const char *service_type, *arg_name;
         SoupSoapMessage *msg;
-        GUPnPServiceProxyAction *ret;
-        GUPnPContext *context;
-        SoupSession *session;
-
-        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), NULL);
-        g_return_val_if_fail (action, NULL);
-        g_return_val_if_fail (callback, NULL);
+        char *control_url, *lang, *full_action;
+        const char *service_type;
 
         /* Create message */
         control_url = gupnp_service_info_get_control_url
@@ -542,6 +567,8 @@ gupnp_service_proxy_begin_action_valist
                              GUPNP_SERVER_ERROR_OTHER,
                              "No service type defined");
 
+                g_object_unref (msg);
+
                 return NULL;
         }
 
@@ -561,64 +588,72 @@ gupnp_service_proxy_begin_action_valist
                                          "u",
                                          service_type);
 
-        /* Arguments */
-        arg_name = va_arg (var_args, const char *);
-        while (arg_name) {
-                GType arg_type;
-                GValue value = { 0, }, transformed_value = { 0, };
-                char *collect_error = NULL;
-                const char *str;
+        return msg;
+}
 
-                arg_type = va_arg (var_args, GType);
-                g_value_init (&value, arg_type);
+/* Received response to action message */
+static void
+action_got_response (SoupMessage             *msg,
+                     GUPnPServiceProxyAction *action)
+{
+        const char *full_action;
+        GUPnPContext *context;
+        SoupSession *session;
 
-                G_VALUE_COLLECT (&value, var_args, 0, &collect_error);
-                if (collect_error) {
-                        g_warning ("Error collecting value: %s\n",
-                                   collect_error);
+        switch (msg->status_code) {
+        case SOUP_STATUS_CANCELLED:
+                /* Silently return */
+                break;
 
-                        g_free (collect_error);
-                        
-                        /* we purposely leak the value here, it might not be
-	                 * in a sane state if an error condition occoured
-	                 */
+        case SOUP_STATUS_METHOD_NOT_ALLOWED:
+                /* Retry with M-POST */
+                msg->method = "M-POST";
 
-                        g_object_unref (msg);
+                soup_message_add_header
+                        (msg->request_headers,
+                         "Man",
+                         "\"http://schemas.xmlsoap.org/soap/envelope/\"; ns=s");
 
-                        return NULL;
-                }
+                /* Rename "SOAPAction" to "s-SOAPAction" */
+                full_action = soup_message_get_header (msg->request_headers,
+                                                       "SOAPAction");
+                soup_message_add_header (msg->request_headers,
+                                         "s-SOAPAction",
+                                         full_action);
+                soup_message_remove_header (msg->request_headers,
+                                            "SOAPAction");
 
-                g_value_init (&transformed_value, G_TYPE_STRING);
-                if (!g_value_transform (&value, &transformed_value)) {
-                        g_warning ("Failed to transform value of type %s "
-                                   "to a string", g_type_name (arg_type));
-                        
-                        g_value_unset (&value);
-                        g_value_unset (&transformed_value);
+                /* And re-queue */
+                context = gupnp_service_info_get_context
+                                (GUPNP_SERVICE_INFO (action->proxy));
+                session = _gupnp_context_get_session (context);
 
-                        g_object_unref (msg);
+                soup_session_requeue_message (session, msg);
 
-                        return NULL;
-                }
+                break;
 
-                soup_soap_message_start_element (msg,
-                                                 arg_name,
-                                                 NULL,
-                                                 NULL);
+        default:
+                /* Success: Call callback */
+                action->callback (action->proxy, action, action->user_data);
 
-                str = g_value_get_string (&transformed_value);
-                soup_soap_message_write_string (msg, str); 
-
-                soup_soap_message_end_element (msg);
-
-                g_value_unset (&value);
-                g_value_unset (&transformed_value);
-                
-                arg_name = va_arg (var_args, const char *);
+                break;
         }
+}
 
+/* Finishes an action message and sends it off */
+static GUPnPServiceProxyAction *
+finish_action_msg (GUPnPServiceProxy              *proxy,
+                   SoupSoapMessage                *msg,
+                   GUPnPServiceProxyActionCallback callback,
+                   gpointer                        user_data)
+{
+        GUPnPServiceProxyAction *ret;
+        GUPnPContext *context;
+        SoupSession *session;
+
+        /* Finish message */
         soup_soap_message_end_element (msg);
-        
+
 	soup_soap_message_end_body (msg);
 	soup_soap_message_end_envelope (msg);
         soup_soap_message_persist (msg);
@@ -628,15 +663,16 @@ gupnp_service_proxy_begin_action_valist
 
         ret->proxy = proxy;
 
-        ret->msg = msg;
+        ret->msg   = msg;
 
-        ret->callback = callback;
+        ret->callback  = callback;
         ret->user_data = user_data;
 
         proxy->priv->pending_actions =
                 g_list_prepend (proxy->priv->pending_actions, ret);
 
-        /* We keep our own reference to the message as well */
+        /* We need to keep our own reference to the message as well,
+         * in order for send_action() to work. */
         g_object_ref (msg);
 
         /* Send the message */
@@ -645,16 +681,153 @@ gupnp_service_proxy_begin_action_valist
 
         soup_session_queue_message (session,
                                     SOUP_MESSAGE (msg),
-                                    (SoupMessageCallbackFn)
-                                        action_got_response,
+                                    (SoupMessageCallbackFn) action_got_response,
                                     ret);
 
+        return ret;
+}
+
+/* Writes a parameter name and GValue pair to @msg */
+static void
+write_in_parameter (const char      *arg_name,
+                    GValue          *value,
+                    SoupSoapMessage *msg)
+{
+        char *str;
+
+        soup_soap_message_start_element (msg,
+                                         arg_name,
+                                         NULL,
+                                         NULL);
+
+        str = gvalue_util_value_get_string (value);
+        soup_soap_message_write_string (msg, str);
+        g_free (str);
+
+        soup_soap_message_end_element (msg);
+}
+
+/**
+ * gupnp_service_proxy_begin_action_valist
+ * @proxy: A #GUPnPServiceProxy
+ * @action: An action
+ * @callback: The callback to call when sending the action has succeeded
+ * or failed
+ * @user_data: User data for @callback
+ * @error: The location where to store any error, or NULL
+ * @var_args: A va_list of tuples of in parameter name, in paramater type, and
+ * in parameter value
+ *
+ * See gupnp_service_proxy_begin_action(); this version takes a va_list for
+ * use by language bindings.
+ *
+ * Return value: A #GUPnPServiceProxyAction handle, or NULL on error. This will
+ * be freed automatically on @callback calling.
+ **/
+GUPnPServiceProxyAction *
+gupnp_service_proxy_begin_action_valist
+                                   (GUPnPServiceProxy              *proxy,
+                                    const char                     *action,
+                                    GUPnPServiceProxyActionCallback callback,
+                                    gpointer                        user_data,
+                                    GError                        **error,
+                                    va_list                         var_args)
+{
+        const char *arg_name;
+        SoupSoapMessage *msg;
+        GUPnPServiceProxyAction *ret;
+
+        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), NULL);
+        g_return_val_if_fail (action, NULL);
+        g_return_val_if_fail (callback, NULL);
+
+        /* Create message */
+        msg = begin_action_msg (proxy, action, error);
+        if (msg == NULL)
+                return NULL;
+
+        /* Arguments */
+        arg_name = va_arg (var_args, const char *);
+        while (arg_name) {
+                GType arg_type;
+                GValue value = { 0, };
+                char *collect_error = NULL;
+
+                arg_type = va_arg (var_args, GType);
+                g_value_init (&value, arg_type);
+
+                G_VALUE_COLLECT (&value, var_args, 0, &collect_error);
+                if (!collect_error) {
+                        write_in_parameter (arg_name, &value, msg);
+
+                        g_value_unset (&value);
+
+                } else {
+                        g_warning ("Error collecting value: %s\n",
+                                   collect_error);
+
+                        g_free (collect_error);
+
+                        /* we purposely leak the value here, it might not be
+	                 * in a sane state if an error condition occoured
+	                 */
+                }
+
+                arg_name = va_arg (var_args, const char *);
+        }
+
+        /* Finish and send off */
+        ret = finish_action_msg (proxy, msg, callback, user_data);
+
         /* Save the current position in the va_list for send_action_valist() */
-        ret->var_args = var_args;
+        G_VA_COPY (ret->var_args, var_args);
 
         return ret;
 }
 
+/**
+ * gupnp_service_proxy_begin_action_hash
+ * @proxy: A #GUPnPServiceProxy
+ * @action: An action
+ * @callback: The callback to call when sending the action has succeeded
+ * or failed
+ * @user_data: User data for @callback
+ * @error: The location where to store any error, or NULL
+ * @hash: A #GHashTable of in parameter name and #GValue pairs
+ *
+ * See gupnp_service_proxy_begin_action(); this version takes a #GHashTable
+ * for runtime generated parameter lists.
+ *
+ * Return value: A #GUPnPServiceProxyAction handle, or NULL on error. This will
+ * be freed automatically on @callback calling.
+ **/
+GUPnPServiceProxyAction *
+gupnp_service_proxy_begin_action_hash
+                                   (GUPnPServiceProxy              *proxy,
+                                    const char                     *action,
+                                    GUPnPServiceProxyActionCallback callback,
+                                    gpointer                        user_data,
+                                    GError                        **error,
+                                    GHashTable                     *hash)
+{
+        SoupSoapMessage *msg;
+
+        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), NULL);
+        g_return_val_if_fail (action, NULL);
+        g_return_val_if_fail (callback, NULL);
+
+        /* Create message */
+        msg = begin_action_msg (proxy, action, error);
+        if (msg == NULL)
+                return NULL;
+
+        /* Arguments */
+        g_hash_table_foreach (hash, (GHFunc) write_in_parameter, msg);
+
+        /* Finish and send off */
+        return finish_action_msg (proxy, msg, callback, user_data);
+}
+
 /**
  * gupnp_service_proxy_end_action
  * @proxy: A #GUPnPServiceProxy
@@ -664,7 +837,7 @@ gupnp_service_proxy_begin_action_valist
  * value location, terminated with NULL. The out parameter values should be
  * freed after use
  *
- * Retrieves the result of @action. The out parameters in @Varargs will be 
+ * Retrieves the result of @action. The out parameters in @Varargs will be
  * filled in, and if an error occurred, @error will be set. In case of
  * a UPnPError the error code will be the same in @error.
  *
@@ -689,35 +862,18 @@ gupnp_service_proxy_end_action (GUPnPServiceProxy       *proxy,
         return ret;
 }
 
-/**
- * gupnp_service_proxy_end_action_valist
- * @proxy: A #GUPnPServiceProxy
- * @action: A #GUPnPServiceProxyAction handle
- * @error: The location where to store any error, or NULL
- * @var_args: A va_list of tuples of out parameter name, out paramater type, 
- * and out parameter value location. The out parameter values should be
- * freed after use
- *
- * See gupnp_service_proxy_end_action(); this version takes a va_list for
- * use by language bindings.
- *
- * Return value: TRUE on success.
- **/
-gboolean
-gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
-                                       GUPnPServiceProxyAction *action,
-                                       GError                 **error,
-                                       va_list                  var_args)
+/* Checks an action response for errors and returns the parsed
+ * SoupSoapResponse object. */
+static SoupSoapResponse *
+check_action_response (GUPnPServiceProxy       *proxy,
+                       GUPnPServiceProxyAction *action,
+                       GError                 **error)
 {
         SoupMessage *soup_msg;
         SoupSoapResponse *response;
         SoupSoapParameter *param;
-        const char *arg_name;
         int code;
 
-        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), FALSE);
-        g_return_val_if_fail (action, FALSE);
-
         soup_msg = SOUP_MESSAGE (action->msg);
 
         /* Check for errors */
@@ -727,10 +883,10 @@ gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
                 break;
         default:
                 set_server_error (error, soup_msg);
-                
+
                 gupnp_service_proxy_action_free (action);
 
-                return FALSE;
+                return NULL;
         }
 
         /* Parse response */
@@ -748,10 +904,10 @@ gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
                                      GUPNP_SERVER_ERROR_INTERNAL_SERVER_ERROR,
                                      soup_msg->reason_phrase);
                 }
-                
+
                 gupnp_service_proxy_action_free (action);
 
-                return FALSE;
+                return NULL;
 	}
 
         /* Check whether we have a Fault */
@@ -776,7 +932,7 @@ gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
 
                         g_object_unref (response);
 
-                        return FALSE;
+                        return NULL;
                 }
 
                 /* Code */
@@ -794,7 +950,7 @@ gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
 
                         g_object_unref (response);
 
-                        return FALSE;
+                        return NULL;
                 }
 
                 /* Description */
@@ -811,101 +967,92 @@ gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
                                    desc);
 
                 g_free (desc);
-                
+
                 /* Cleanup */
                 gupnp_service_proxy_action_free (action);
 
                 g_object_unref (response);
 
-                return FALSE;
+                return NULL;
         }
 
+        return response;
+}
+
+/* Reads a value into the parameter name and initialised GValue pair
+ * from @response */
+static void
+read_out_parameter (const char       *arg_name,
+                    GValue           *value,
+                    SoupSoapResponse *response)
+{
+        SoupSoapParameter *param;
+        char *str;
+
+        /* Try to find a matching paramater in the response */
+        param = soup_soap_response_get_first_parameter_by_name
+                                                (response, arg_name);
+        if (!param) {
+                g_warning ("Could not find variable \"%s\" in response",
+                           arg_name);
+
+                return;
+        }
+
+        /* Parse into @value */
+        str = soup_soap_parameter_get_string_value (param);
+
+        gvalue_util_set_value_from_string (value, str);
+
+        g_free (str);
+
+        return;
+}
+
+/**
+ * gupnp_service_proxy_end_action_valist
+ * @proxy: A #GUPnPServiceProxy
+ * @action: A #GUPnPServiceProxyAction handle
+ * @error: The location where to store any error, or NULL
+ * @var_args: A va_list of tuples of out parameter name, out paramater type,
+ * and out parameter value location. The out parameter values should be
+ * freed after use
+ *
+ * See gupnp_service_proxy_end_action(); this version takes a va_list for
+ * use by language bindings.
+ *
+ * Return value: TRUE on success.
+ **/
+gboolean
+gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
+                                       GUPnPServiceProxyAction *action,
+                                       GError                 **error,
+                                       va_list                  var_args)
+{
+        SoupSoapResponse *response;
+        const char *arg_name;
+
+        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), FALSE);
+        g_return_val_if_fail (action, FALSE);
+
+        /* Check response for errors and do initial parsing */
+        response = check_action_response (proxy, action, error);
+        if (response == NULL)
+                return FALSE;
+
         /* Arguments */
         arg_name = va_arg (var_args, const char *);
         while (arg_name) {
                 GType arg_type;
-                GValue value = { 0, }, tmp_value = { 0, };
+                GValue value = { 0, };
                 char *copy_error = NULL;
-                char *str;
-                int i;
-
-                param = soup_soap_response_get_first_parameter_by_name
-                                                        (response, arg_name);
-                if (!param) {
-                        g_set_error (error,
-                                     GUPNP_CONTROL_ERROR,
-                                     GUPNP_CONTROL_ERROR_INVALID_ARGS,
-                                     "Could not find variable \"%s\" in "
-                                     "response",
-                                     arg_name);
-
-                        gupnp_service_proxy_action_free (action);
-
-                        g_object_unref (response);
-
-                        return FALSE;
-                }
 
                 arg_type = va_arg (var_args, GType);
 
                 g_value_init (&value, arg_type);
 
-                switch (arg_type) {
-                case G_TYPE_STRING:
-                        str = soup_soap_parameter_get_string_value (param);
+                read_out_parameter (arg_name, &value, response);
 
-                        g_value_set_string_take_ownership (&value, str);
-
-                        break;
-                case G_TYPE_INT:
-                        i = soup_soap_parameter_get_int_value (param);
-
-                        g_value_set_int (&value, i);
-
-                        break;
-                default:
-                        /* Try to convert */
-                        if (g_value_type_transformable (G_TYPE_STRING,
-                                                        arg_type)) {
-                                str = soup_soap_parameter_get_string_value
-                                                                        (param);
-
-                                g_value_init (&tmp_value, G_TYPE_STRING);
-                                g_value_set_string_take_ownership (&tmp_value,
-                                                                   str);
-
-                                g_value_transform (&tmp_value, &value);
-
-                                g_value_unset (&tmp_value);
-
-                        } else if (g_value_type_transformable (G_TYPE_INT,
-                                                               arg_type)) {
-                                i = soup_soap_parameter_get_int_value (param);
-
-                                g_value_init (&tmp_value, G_TYPE_INT);
-                                g_value_set_int (&tmp_value, i);
-
-                                g_value_transform (&tmp_value, &value);
-
-                                g_value_unset (&tmp_value);
-
-                        } else {
-                                g_warning ("Failed to transform integer "
-                                           "value to type %s",
-                                           g_type_name (arg_type));
-
-                                g_value_unset (&value);
-
-                                gupnp_service_proxy_action_free (action);
-
-                                g_object_unref (response);
-
-                                return FALSE;
-                        }
-
-                        break;
-                }
-                
                 G_VALUE_LCOPY (&value, var_args, 0, &copy_error);
 
                 g_value_unset (&value);
@@ -914,12 +1061,6 @@ gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
                         g_warning ("Error copying value: %s", copy_error);
 
                         g_free (copy_error);
-                        
-                        gupnp_service_proxy_action_free (action);
-
-                        g_object_unref (response);
-
-                        return FALSE;
                 }
 
                 arg_name = va_arg (var_args, const char *);
@@ -933,6 +1074,45 @@ gupnp_service_proxy_end_action_valist (GUPnPServiceProxy       *proxy,
         return TRUE;
 }
 
+/**
+ * gupnp_service_proxy_end_action_hash
+ * @proxy: A #GUPnPServiceProxy
+ * @action: A #GUPnPServiceProxyAction handle
+ * @error: The location where to store any error, or NULL
+ * @hash: A #GHashTable of out parameter name and initialised #GValue pairs
+ *
+ * See gupnp_service_proxy_end_action(); this version takes a #GHashTable
+ * for runtime generated parameter lists.
+ *
+ * Return value: TRUE on success.
+ **/
+gboolean
+gupnp_service_proxy_end_action_hash (GUPnPServiceProxy       *proxy,
+                                     GUPnPServiceProxyAction *action,
+                                     GError                 **error,
+                                     GHashTable              *hash)
+{
+        SoupSoapResponse *response;
+
+        g_return_val_if_fail (GUPNP_IS_SERVICE_PROXY (proxy), FALSE);
+        g_return_val_if_fail (action, FALSE);
+
+        /* Check response for errors and do initial parsing */
+        response = check_action_response (proxy, action, error);
+        if (response == NULL)
+                return FALSE;
+
+        /* Read arguments */
+        g_hash_table_foreach (hash, (GHFunc) read_out_parameter, response);
+
+        /* Cleanup */
+        gupnp_service_proxy_action_free (action);
+
+        g_object_unref (response);
+
+        return TRUE;
+}
+
 /**
  * gupnp_service_proxy_cancel_action
  * @proxy: A #GUPnPServiceProxy
@@ -952,7 +1132,7 @@ gupnp_service_proxy_cancel_action (GUPnPServiceProxy       *proxy,
 
         context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (proxy));
         session = _gupnp_context_get_session (context);
-        
+
         soup_message_set_status (SOUP_MESSAGE (action->msg),
                                  SOUP_STATUS_CANCELLED);
         soup_session_cancel_message (session, SOUP_MESSAGE (action->msg));
@@ -968,7 +1148,7 @@ gupnp_service_proxy_cancel_action (GUPnPServiceProxy       *proxy,
  * @callback: The callback to call when @variable changes
  * @user_data: User data for @callback
  *
- * Sets up @callback to be called whenever a change notification for 
+ * Sets up @callback to be called whenever a change notification for
  * @variable is recieved.
  *
  * Return value: TRUE on success.
@@ -1173,7 +1353,7 @@ server_handler (SoupServerContext *server_context,
         seq = atoi (hdr);
         if (seq > proxy->priv->seq) {
                 /* Oops, we missed a notify. Resubscribe .. */
-                unsubscribe (proxy, FALSE);
+                unsubscribe (proxy);
                 subscribe (proxy);
 
                 /* Message was OK otherwise */
@@ -1323,6 +1503,9 @@ subscription_expire (gpointer user_data)
         g_free (timeout);
 
         /* And send it off */
+        proxy->priv->pending_messages = 
+                g_list_prepend (proxy->priv->pending_messages, msg);
+
         session = _gupnp_context_get_session (context);
 
         soup_session_queue_message (session,
@@ -1343,6 +1526,14 @@ subscribe_got_response (SoupMessage       *msg,
 {
         GError *error;
 
+        /* Cancelled? */
+        if (msg->status_code == SOUP_STATUS_CANCELLED)
+                return;
+
+        /* Remove from pending messages list */
+        proxy->priv->pending_messages = 
+                g_list_remove (proxy->priv->pending_messages, msg);
+
         /* Check whether the subscription is still wanted */
         if (!proxy->priv->subscribed)
                 return;
@@ -1403,8 +1594,7 @@ subscribe_got_response (SoupMessage       *msg,
                 /* Subscription failed. */
                 error = g_error_new (GUPNP_EVENTING_ERROR,
                                      GUPNP_EVENTING_ERROR_SUBSCRIPTION_FAILED,
-                                     "(Re)subscription failed: %d",
-                                     msg->status_code);
+                                     msg->reason_phrase);
 
  ERROR:
                 proxy->priv->subscribed = FALSE;
@@ -1508,6 +1698,9 @@ subscribe (GUPnPServiceProxy *proxy)
                                  proxy);
 
         /* And send our subscription message off */
+        proxy->priv->pending_messages = 
+                g_list_prepend (proxy->priv->pending_messages, msg);
+
         session = _gupnp_context_get_session (context);
 
         soup_session_queue_message (session,
@@ -1521,7 +1714,7 @@ subscribe (GUPnPServiceProxy *proxy)
  * Unsubscribe from this service.
  **/
 static void
-unsubscribe (GUPnPServiceProxy *proxy, gboolean sync)
+unsubscribe (GUPnPServiceProxy *proxy)
 {
         GUPnPContext *context;
         SoupMessage *msg;
@@ -1549,14 +1742,11 @@ unsubscribe (GUPnPServiceProxy *proxy, gboolean sync)
                                  "SID",
                                  proxy->priv->sid);
 
-        /* And send it off */
+        /* And queue it */
         session = _gupnp_context_get_session (context);
 
-        if (sync)
-                soup_session_send_message (session, msg);
-        else
-                soup_session_queue_message (session, msg, NULL, NULL);
-        
+        soup_session_queue_message (session, msg, NULL, NULL);
+
         /* Reset SID */
         g_free (proxy->priv->sid);
         proxy->priv->sid = NULL;
@@ -1565,24 +1755,16 @@ unsubscribe (GUPnPServiceProxy *proxy, gboolean sync)
         proxy->priv->seq = 0;
 
         /* Remove subscription timeout */
-        g_source_remove (proxy->priv->subscription_timeout_id);
-        proxy->priv->subscription_timeout_id = 0;
+        if (proxy->priv->subscription_timeout_id) {
+                g_source_remove (proxy->priv->subscription_timeout_id);
+                proxy->priv->subscription_timeout_id = 0;
+        }
 
         /* Remove server handler */
         server = _gupnp_context_get_server (context);
         soup_server_remove_handler (server, proxy->priv->path);
 }
 
-void
-_gupnp_service_proxy_unsubscribe_sync (GUPnPServiceProxy *proxy)
-{
-        if (proxy->priv->subscribed) {
-                unsubscribe (proxy, TRUE);
-
-                proxy->priv->subscribed = FALSE;
-        }
-}
-
 /**
  * gupnp_service_proxy_set_subscribed
  * @proxy: A #GUPnPServiceProxy
@@ -1609,7 +1791,7 @@ gupnp_service_proxy_set_subscribed (GUPnPServiceProxy *proxy,
         if (subscribed)
                 subscribe (proxy);
         else
-                unsubscribe (proxy, FALSE);
+                unsubscribe (proxy);
 
         g_object_notify (G_OBJECT (proxy), "subscribed");
 }
@@ -1642,6 +1824,7 @@ gupnp_service_proxy_get_subscribed (GUPnPServiceProxy *proxy)
  **/
 GUPnPServiceProxy *
 _gupnp_service_proxy_new (GUPnPContext  *context,
+                          XmlDocWrapper *doc,
                           xmlNode       *element,
                           const char    *udn,
                           const char    *service_type,
@@ -1651,6 +1834,7 @@ _gupnp_service_proxy_new (GUPnPContext  *context,
         GUPnPServiceProxy *proxy;
 
         g_return_val_if_fail (GUPNP_IS_CONTEXT (context), NULL);
+        g_return_val_if_fail (IS_XML_DOC_WRAPPER (doc), NULL);
         g_return_val_if_fail (element != NULL, NULL);
         g_return_val_if_fail (location != NULL, NULL);
         g_return_val_if_fail (url_base != NULL, NULL);
@@ -1661,6 +1845,7 @@ _gupnp_service_proxy_new (GUPnPContext  *context,
                               "udn", udn,
                               "service-type", service_type,
                               "url-base", url_base,
+                              "document", doc,
                               "element", element,
                               NULL);
 
